@@ -67,6 +67,10 @@ except ImportError:  # pragma: no cover
     from coverage.collector import FileDisposition
 
 
+running_under_pytest = False
+global_force_seed = None
+
+
 def new_random():
     import random
     return random.Random(random.getrandbits(128))
@@ -91,7 +95,7 @@ class Example(object):
 
 
 def example(*args, **kwargs):
-    """A decorator to that ensures a specific example is always tested."""
+    """A decorator which ensures a specific example is always tested."""
     if args and kwargs:
         raise InvalidArgument(
             'Cannot mix positional and keyword arguments for examples'
@@ -112,17 +116,12 @@ def example(*args, **kwargs):
 def reify_and_execute(
     search_strategy, test,
     print_example=False,
-    is_final=False,
+    is_final=False, collector=None
 ):
     def run(data):
         with BuildContext(data, is_final=is_final):
-            orig = sys.gettrace()
-            try:  # pragma: no cover
-                sys.settrace(None)
-                import random as rnd_module
-                rnd_module.seed(0)
-            finally:  # pragma: no cover
-                sys.settrace(orig)
+            import random as rnd_module
+            rnd_module.seed(0)
             args, kwargs = data.draw(search_strategy)
 
             if print_example:
@@ -133,7 +132,15 @@ def reify_and_execute(
                 report(
                     lambda: 'Trying example: %s(%s)' % (
                         test.__name__, arg_string(test, args, kwargs)))
-            return test(*args, **kwargs)
+            if collector is None:
+                return test(*args, **kwargs)
+            else:  # pragma: no cover
+                try:
+                    collector.start()
+                    return test(*args, **kwargs)
+                finally:
+                    collector.stop()
+
     return run
 
 
@@ -419,14 +426,20 @@ def perform_health_checks(random, settings, test_runner, search_strategy):
 
 def get_random_for_wrapped_test(test, wrapped_test):
     settings = wrapped_test._hypothesis_internal_use_settings
+    wrapped_test._hypothesis_internal_use_generated_seed = None
 
     if wrapped_test._hypothesis_internal_use_seed is not None:
         return Random(
             wrapped_test._hypothesis_internal_use_seed)
     elif settings.derandomize:
         return Random(function_digest(test))
+    elif global_force_seed is not None:
+        return Random(global_force_seed)
     else:
-        return new_random()
+        import random
+        seed = random.getrandbits(128)
+        wrapped_test._hypothesis_internal_use_generated_seed = seed
+        return Random(seed)
 
 
 def process_arguments_to_given(
@@ -525,25 +538,11 @@ STDLIB = os.path.dirname(os.__file__)
 
 
 def hypothesis_check_include(filename):  # pragma: no cover
+    if is_hypothesis_file(filename):
+        return False
     return filename.endswith('.py')
 
 
-def hypothesis_should_trace(original_filename, frame):  # pragma: no cover
-    disp = FileDisposition()
-    assert original_filename is not None
-    disp.original_filename = original_filename
-    disp.canonical_filename = encoded_filepath(
-        canonical_filename(original_filename))
-    disp.source_filename = disp.canonical_filename
-    disp.reason = ''
-    disp.file_tracer = None
-    disp.has_dynamic_filename = False
-    disp.trace = hypothesis_check_include(disp.canonical_filename)
-    if not disp.trace:
-        disp.reason = 'hypothesis internal reasons'
-    return disp
-
-
 def escalate_warning(msg, slug=None):  # pragma: no cover
     if slug is not None:
         msg = '%s (%s)' % (msg, slug)
@@ -552,17 +551,26 @@ def escalate_warning(msg, slug=None):  # pragma: no cover
     )
 
 
-@attr.s(slots=True, frozen=True)
-class Line(object):
-    filename = attr.ib()
-    lineno = attr.ib()
-
-
-@attr.s(slots=True, frozen=True)
 class Arc(object):
-    filename = attr.ib()
-    source = attr.ib()
-    target = attr.ib()
+    __slots__ = ('filename', 'source', 'target')
+
+    def __init__(self, filename, source, target):
+        self.filename = filename
+        self.source = source
+        self.target = target
+
+
+ARC_CACHE = {}
+
+
+def arc(filename, source, target):
+    try:
+        return ARC_CACHE[filename][source][target]
+    except KeyError:
+        result = Arc(filename, source, target)
+        ARC_CACHE.setdefault(
+            filename, {}).setdefault(source, {})[target] = result
+        return result
 
 
 in_given = False
@@ -583,14 +591,20 @@ class StateForActualGivenExecution(object):
         self.__was_flaky = False
         self.random = random
         self.__warned_deadline = False
+        self.__existing_collector = None
+        self.__test_runtime = None
+        self.__in_final_replay = False
+
         if self.settings.deadline is None:
             self.test = test
         else:
             @proxies(test)
             def timed_test(*args, **kwargs):
+                self.__test_runtime = None
                 start = time.time()
                 result = test(*args, **kwargs)
                 runtime = (time.time() - start) * 1000
+                self.__test_runtime = runtime
                 if self.settings.deadline is not_set:
                     if (
                         not self.__warned_deadline and
@@ -606,15 +620,13 @@ class StateForActualGivenExecution(object):
                             'it to None to disable this check entirely.') % (
                                 runtime, ceil(runtime / 100) * 100,
                         ))
-                elif runtime >= self.settings.deadline:
-                    raise DeadlineExceeded((
-                        'Test took %.2fms, which exceeds the deadline of '
-                        '%.2fms') % (runtime, self.settings.deadline)
-                    )
+                elif runtime >= self.current_deadline:
+                    raise DeadlineExceeded(runtime, self.settings.deadline)
                 return result
             self.test = timed_test
 
         self.coverage_data = CoverageData()
+        self.files_to_propagate = set()
 
         if settings.use_coverage and not IN_COVERAGE_TESTS:  # pragma: no cover
             if Collector._collectors:
@@ -623,7 +635,7 @@ class StateForActualGivenExecution(object):
             self.collector = Collector(
                 branch=True,
                 timid=FORCE_PURE_TRACER,
-                should_trace=hypothesis_should_trace,
+                should_trace=self.should_trace,
                 check_include=hypothesis_check_include,
                 concurrency='thread',
                 warn=escalate_warning,
@@ -637,7 +649,36 @@ class StateForActualGivenExecution(object):
         else:
             self.collector = None
 
+    @property
+    def current_deadline(self):
+        base = self.settings.deadline
+        if self.__in_final_replay:
+            return base
+        else:
+            return base * 1.25
+
+    def should_trace(self, original_filename, frame):  # pragma: no cover
+        disp = FileDisposition()
+        assert original_filename is not None
+        disp.original_filename = original_filename
+        disp.canonical_filename = encoded_filepath(
+            canonical_filename(original_filename))
+        disp.source_filename = disp.canonical_filename
+        disp.reason = ''
+        disp.file_tracer = None
+        disp.has_dynamic_filename = False
+        disp.trace = hypothesis_check_include(disp.canonical_filename)
+        if not disp.trace:
+            disp.reason = 'hypothesis internal reasons'
+        elif self.__existing_collector is not None:
+            check = self.__existing_collector.should_trace(
+                original_filename, frame)
+            if check.trace:
+                self.files_to_propagate.add(check.canonical_filename)
+        return disp
+
     def hijack_collector(self, collector):  # pragma: no cover
+        self.__existing_collector = collector
         original_save_data = collector.save_data
 
         def save_data(covdata):
@@ -647,14 +688,14 @@ class StateForActualGivenExecution(object):
                     filename: {
                         arc: None
                         for arc in self.coverage_data.arcs(filename)}
-                    for filename in self.coverage_data.measured_files()
+                    for filename in self.files_to_propagate
                 })
             else:
                 covdata.add_lines({
                     filename: {
                         line: None
                         for line in self.coverage_data.lines(filename)}
-                    for filename in self.coverage_data.measured_files()
+                    for filename in self.files_to_propagate
                 })
             collector.save_data = original_save_data
         collector.save_data = save_data
@@ -680,14 +721,11 @@ class StateForActualGivenExecution(object):
                 original = sys.gettrace()
                 sys.settrace(None)
                 try:
-                    try:
-                        self.collector.data = {}
-                        self.collector.start()
-                        result = self.test_runner(data, reify_and_execute(
-                            self.search_strategy, self.test,
-                        ))
-                    finally:
-                        self.collector.stop()
+                    self.collector.data = {}
+                    result = self.test_runner(data, reify_and_execute(
+                        self.search_strategy, self.test,
+                        collector=self.collector
+                    ))
                 finally:
                     sys.settrace(original)
                     covdata = CoverageData()
@@ -696,10 +734,10 @@ class StateForActualGivenExecution(object):
                     for filename in covdata.measured_files():
                         if is_hypothesis_file(filename):
                             continue
-                        for lineno in covdata.lines(filename):
-                            data.add_tag(Line(filename, lineno))
-                        for source, target in covdata.arcs(filename):
-                            data.add_tag(Arc(filename, source, target))
+                        data.tags.update(
+                            arc(filename, source, target)
+                            for source, target in covdata.arcs(filename)
+                        )
             if result is not None and self.settings.perform_health_check:
                 fail_health_check(self.settings, (
                     'Tests run under @given should return None, but '
@@ -714,10 +752,11 @@ class StateForActualGivenExecution(object):
             StopTest,
         ) + exceptions_to_reraise:
             raise
-        except Exception:
+        except Exception as e:
             escalate_hypothesis_internal_error()
-            data.__expected_exception = traceback.format_exc()
-            verbose_report(data.__expected_exception)
+            data.__expected_traceback = traceback.format_exc()
+            data.__expected_exception = e
+            verbose_report(data.__expected_traceback)
 
             error_class, _, tb = sys.exc_info()
 
@@ -804,6 +843,8 @@ class StateForActualGivenExecution(object):
 
         flaky = 0
 
+        self.__in_final_replay = True
+
         for falsifying_example in self.falsifying_examples:
             self.__was_flaky = False
             raised_exception = False
@@ -828,10 +869,28 @@ class StateForActualGivenExecution(object):
                 report(traceback.format_exc())
 
             if not raised_exception:
-                report(
-                    'Failed to reproduce exception. Expected: \n' +
-                    falsifying_example.__expected_exception,
-                )
+                if (
+                    isinstance(
+                        falsifying_example.__expected_exception,
+                        DeadlineExceeded
+                    ) and self.__test_runtime is not None
+                ):
+                    report((
+                        'Unreliable test timings! On an initial run, this '
+                        'test took %.2fms, which exceeded the deadline of '
+                        '%.2fms, but on a subsequent run it took %.2f ms, '
+                        'which did not. If you expect this sort of '
+                        'variability in your test timings, consider turning '
+                        'deadlines off for this test by setting deadline=None.'
+                    ) % (
+                        falsifying_example.__expected_exception.runtime,
+                        self.settings.deadline, self.__test_runtime
+                    ))
+                else:
+                    report(
+                        'Failed to reproduce exception. Expected: \n' +
+                        falsifying_example.__expected_traceback,
+                    )
 
                 filter_message = (
                     'Unreliable test data: Failed to reproduce a failure '
@@ -918,6 +977,17 @@ def given(*given_arguments, **given_kwargs):
             # Tell pytest to omit the body of this function from tracebacks
             __tracebackhide__ = True
 
+            if getattr(test, 'is_hypothesis_test', False):
+                note_deprecation((
+                    'You have applied @given to a test more than once. In '
+                    'future this will be an error. Applying @given twice '
+                    'wraps the test twice, which can be extremely slow. A '
+                    'similar effect can be gained by combining the arguments '
+                    'to the two calls to given. For example, instead of '
+                    '@given(booleans()) @given(integers()), you could write '
+                    '@given(booleans(), integers())'
+                ))
+
             settings = wrapped_test._hypothesis_internal_use_settings
 
             random = get_random_for_wrapped_test(test, wrapped_test)
@@ -951,12 +1021,28 @@ def given(*given_arguments, **given_kwargs):
             ):
                 return
 
-            perform_health_checks(
-                random, settings, test_runner, search_strategy)
+            try:
+                perform_health_checks(
+                    random, settings, test_runner, search_strategy)
 
-            state = StateForActualGivenExecution(
-                test_runner, search_strategy, test, settings, random)
-            state.run()
+                state = StateForActualGivenExecution(
+                    test_runner, search_strategy, test, settings, random)
+                state.run()
+            except:
+                generated_seed = \
+                    wrapped_test._hypothesis_internal_use_generated_seed
+                if generated_seed is not None:
+                    if running_under_pytest:
+                        report((
+                            'You can add @seed(%(seed)d) to this test or run '
+                            'pytest with --hypothesis-seed=%(seed)d to '
+                            'reproduce this failure.') % {
+                                'seed': generated_seed},)
+                    else:
+                        report((
+                            'You can add @seed(%d) to this test to reproduce '
+                            'this failure.') % (generated_seed,))
+                raise
 
         for attrib in dir(test):
             if not (attrib.startswith('_') or hasattr(wrapped_test, attrib)):
