@@ -26,7 +26,6 @@ import math
 import codecs
 import platform
 import importlib
-from gzip import GzipFile
 from base64 import b64encode
 from decimal import Context, Decimal, Inexact
 from hashlib import sha1
@@ -107,6 +106,9 @@ def quiet_raise(exc):
     def zero_byte_sequence(n):
         return bytes(n)
 
+    def int_to_byte(i):
+        return bytes([i])
+
     import struct
 
     struct_pack = struct.pack
@@ -159,6 +161,8 @@ else:
             raise OverflowError('int too big to convert')
         return hbytes(result)
 
+    int_to_byte = chr
+
     def bytes_from_list(ls):
         return hbytes(bytearray(ls))
 
@@ -239,6 +243,23 @@ else:
     from time import time as benchmark_time
 
 
+# coverage mixes unicode and str filepaths on Python 2, which causes us
+# problems if we're running under unicodenazi (it might also cause problems
+# when not running under unicodenazi, but hard to say for sure). This method
+# exists to work around that: If we're given a unicode filepath, we turn it
+# into a string file path using the appropriate encoding. See
+# https://bitbucket.org/ned/coveragepy/issues/602/ for more information.
+if PY2:
+    def encoded_filepath(filepath):
+        if isinstance(filepath, text_type):
+            return filepath.encode(sys.getfilesystemencoding())
+        else:
+            return filepath
+else:
+    def encoded_filepath(filepath):
+        return filepath
+
+
 def a_good_encoding():
     return 'utf-8'
 
@@ -486,8 +507,6 @@ else:
 # an existing file where you're not allowed to. This is rather less consistent
 # between versions than might be hoped.
 if PY3:
-    # Note: This only works on >= 3.3, but we only support >= 3.3 so that's
-    # fine
     FileExistsError = FileExistsError
 
 elif WINDOWS:
@@ -518,6 +537,12 @@ else:
     ceil = math.ceil
 
 
+try:
+    from math import gcd
+except ImportError:
+    from fractions import gcd
+
+
 if PY2:
     def b64decode(s):
         from base64 import b64decode as base
